//----------------------------------------------------------------------------
//
// File:        dumpcpu.cpp
// Date:        23-Feb-1998
// Programmer:  Marc Rousseau
//
// Description:
//
// Copyright (c) 1994-2000 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <ctype.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.hpp"
#include "cartridge.hpp"
#include "tms9900.hpp"

#define BANK_SIZE	0x1000
#define BANK_MASK	0xF000

#define UNKNOWN		0x0000

#define DATA_BYTE	0x0001
#define DATA_WORD	0x0002
#define DATA_STRING	0x0004
#define DATA_TEXT	0x0008
#define DATA		0x000F

#define DATA_SINGLE	0x0010

#define PRESENT		0x0020

#define CODE		0x0100
#define CODE_START	0x0200

#define LABEL_DATA	0x0800
#define LABEL_JMP	0x1000
#define LABEL_B		0x2000
#define LABEL_BL	0x4000
#define LABEL_BLWP	0x8000
#define LABEL		0xF800

char   LastLabel  [5] = "A@";

char  *Labels     [ 0x10000 ];
USHORT Attributes [ 0x10000 ];
UCHAR  Arguments  [ 0x10000 ];

extern "C" sLookUp LookUp [ 16 ];
extern "C" sOpCode OpCodes [ 69 ];

extern "C" UCHAR  CpuMemory  [ 0x10000 ];

extern USHORT DisassembleASM ( USHORT, UCHAR *, char * );
extern void swab ( char *, char *, unsigned );

FILE *outFile;

int sortFunction ( const sOpCode *p1, const sOpCode *p2 )
{
    if ( p1->opCode < p2->opCode ) return -1;
    if ( p1->opCode > p2->opCode ) return 1;
    return 0;
}

void Init ()
{
    memset ( Attributes, 0, sizeof ( Attributes ));

    // Sort the OpCode table by OpCode
    qsort ( OpCodes, SIZE ( OpCodes ), sizeof ( OpCodes[0] ), ( QSORT_FUNC ) sortFunction );

    // Create the LookUp table using the high 4 bits of each OpCode
    unsigned x, last = 0;
    for ( x = 0; x < SIZE ( LookUp ) - 1; x++ ) {
        int index = last + 1;
        while ( x + 1 > ( unsigned ) ( OpCodes[index].opCode >> 12 )) index++;
        LookUp [x].opCode = &OpCodes [last];
        LookUp [x].size = index - last - 1;
        last = index;
    }
    LookUp [x].opCode = &OpCodes [last];
    LookUp [x].size = SIZE ( OpCodes ) - last - 1;
}

char *GetNextLabel ()
{
    if ( LastLabel [1] == 'Z' ) {
        LastLabel [0]++;
        LastLabel [1] = 'A';
    } else {
        LastLabel [1]++;
    }
    return strdup ( LastLabel );
}

USHORT GetShort ( UCHAR *ptr )
{
    return ( ptr[0] << 8 ) | ptr [1];
}

struct sStackEntry {
    USHORT  address;
};

sStackEntry AddressStack [ 0x1000 ];
unsigned stackTop = 0;

bool PushAddress ( USHORT address, USHORT type, bool bottom = false )
{
    if ( ! bottom ) Attributes [ address ] |= type;
    if ( stackTop >= SIZE ( AddressStack )) {
        fprintf ( stderr, "========= ** ERROR ** Address Stack Overflow ** ERROR ** =========\n" );
        return false;
    }
    if ( Attributes [ address ] & CODE ) return true;
if (( Attributes [ address ] & DATA ) && ( type != LABEL_BLWP )) fprintf ( stderr, "ERROR pushed address marked as DATA\n" );

    sStackEntry *pEntry = AddressStack;

    if ( bottom ) {
        memmove ( AddressStack + 1, AddressStack, sizeof ( USHORT ) * stackTop++ );
    } else {
        pEntry += stackTop++;
    }

    pEntry->address = address;

    return true;
}

bool IsStackEmpty ()
{
    return ( stackTop == 0 ) ? true : false;
}

bool PopAddress ( sStackEntry *entry )
{
    if ( stackTop == 0 ) return false;
    *entry = AddressStack [ --stackTop ];
    return true;
}

void InterpretHeader ( USHORT )
{
}

void TraceCode ()
{
    char buffer [ 80 ];

    sStackEntry top;
    while ( PopAddress ( &top ) == true ) {
        USHORT address = top.address;
fprintf ( stderr, "--- popping address >%04X\n", address );

        if ( Attributes [ address ] & LABEL_BLWP ) {
            Attributes [ address ] |= DATA_WORD;
            Attributes [ address + 2 ] |= DATA_WORD;
            PushAddress ( * ( USHORT * ) &CpuMemory [ address + 2 ], 0 );
            continue;
        }

        for ( EVER ) {

            if (( Attributes [ address ] & PRESENT ) == 0 ) {
fprintf ( stderr, "Error - no code loaded @ >%04X\n", address );
                break;
            }

            // If we've already been here, skip it
            if ( Attributes [ address ] & CODE ) {
                if (( Attributes [ address ] & CODE_START ) == 0 ) {
                    fprintf ( stderr, "Error - code doesn't begin with CODE_START!\n" );
                }
                break;
            }

            USHORT opcode = * ( USHORT * ) &CpuMemory [ address ];

            // Mark all bytes in this instruction as CODE

            USHORT nextAddress = DisassembleASM ( address, &CpuMemory [ address ], buffer );
            int size = nextAddress - address;
            int badCode = 0;

            Attributes [ address ] |= CODE_START;

            for ( int i = 0; ( badCode == 0 ) && ( i < size ); i++ ) { 
                if ( Attributes [ address + i ] & CODE ) {
                    fprintf ( stderr, "Error parsing code - CODE attribute already set (%04X+%d)!\n", address, i );
                    badCode = i;
                    break;
                }
                if ( Attributes [ address + i ] & DATA ) {
                    fprintf ( stderr, "Error parsing code - DATA attribute set (%04X)!\n", address + i );
                    badCode = i + 1;
                    break;
                }
                if ( i && ( Attributes [ address + i ] & LABEL )) {
                    fprintf ( stderr, "Error parsing code - LABEL attribute set (%04X)!\n", address + i );
                    badCode = i + 1;
                    break;
                }
                Attributes [ address + i ] |= CODE;
            }
            if ( badCode ) {
                Attributes [ address ] &= ~CODE_START;
                for ( int i = 0; i < badCode; i++ ) {
                    Attributes [ address + i ] &= ~CODE;
                }
                break;
            }

            // Look for B instructions
            if (( opcode & 0xFFC0 ) == 0x0440 ) {
                if ( opcode == 0x0460 ) {
                    USHORT target = * ( USHORT * ) &CpuMemory [ address + 2 ];
                    PushAddress ( target, LABEL_B );
                }
                break;
            }

            // Look for BL instructions
            if (( opcode & 0xFFC0 ) == 0x0680 ) {
                if ( opcode == 0x06A0 ) {
                    USHORT target = * ( USHORT * ) &CpuMemory [ address + 2 ];
                    PushAddress ( target, LABEL_BL );
                    for ( int i = 0; i < Arguments [ target ]; i++ ) {
                        Attributes [ nextAddress + 2 * i ] |= DATA_WORD;
                        Attributes [ nextAddress + 2 * i + 1 ] |= DATA_WORD;
                    }
                    nextAddress += Arguments [ target ] * 2;
                }
            }

            // Look for BLWP instructions
            if (( opcode & 0xFFC0 ) == 0x0400 ) {
                USHORT target = * ( USHORT * ) &CpuMemory [ address + 2 ];
                PushAddress ( target, LABEL_BLWP );
                for ( int i = 0; i < Arguments [ target ]; i++ ) {
                    Attributes [ nextAddress + 2 * i ] |= DATA_WORD;
                    Attributes [ nextAddress + 2 * i + 1 ] |= DATA_WORD;
                }
                nextAddress += Arguments [ target ] * 2;
            }

            // Look for jump instructions (except NOP)
            if (( opcode > 0x1000 ) && ( opcode < 0x1D00 )) {
                USHORT target = address + * ( char * ) &CpuMemory [ address ] * 2 + 2;
                PushAddress ( target, LABEL_JMP );
                if (( opcode & 0xFF00 ) == 0x1000 ) {
                    break;
                }
            }

            // Look for RTWP instructions
            if ( opcode == 0x0380 ) {
                break;
            }

            address = nextAddress;
        }
    }
}

USHORT DumpByte ( USHORT address, char *buffer )
{
    USHORT limit = ( address & BANK_MASK ) + BANK_SIZE - 1;
    char *tempPtr = buffer + sprintf ( buffer, "BYTE " );
    for ( int i = 0; i < 8; i++ ) {
        if ( Attributes [ address ] & CODE ) {
            fprintf ( stderr, "**ERROR->%04X** %s DATA marked as %s\n", address, "Byte", "CODE" );
        }
        tempPtr += sprintf ( tempPtr, ">%02X", CpuMemory [ address ^ 1 ] );
        if ( Attributes [ address ] & DATA_SINGLE ) break;
        if ( ++address > limit ) break;
        if (( Attributes [ address ] != 0 ) && (( Attributes [ address ] & DATA_BYTE ) == 0 )) break;
        if (( Attributes [ address ] & PRESENT ) == 0 ) break;
        if ( Attributes [ address ] & LABEL ) break;
        if ( i < 7 ) *tempPtr++ = ',';
    }
    *tempPtr = '\0';
    return address;
}

USHORT DumpWord ( USHORT address, char *buffer )
{
    USHORT limit = ( address & BANK_MASK ) + BANK_SIZE - 1;
    char *tempPtr = buffer + sprintf ( buffer, "DATA " );
    for ( int i = 0; i < 8; i += 2 ) {
        if (( Attributes [ address ] & DATA_WORD ) == 0 ) break;
        if ( Attributes [ address ] & CODE ) {
            fprintf ( stderr, "**ERROR->%04X** %s DATA marked as %s\n", address, "Word", "CODE" );
        }
        tempPtr += sprintf ( tempPtr, ">%04X", * ( USHORT * ) &CpuMemory [ address ]);
        if ( Attributes [ address ] & DATA_SINGLE ) break;
        address += 2;
        if ( address > limit ) break;
        if (( Attributes [ address ] & PRESENT ) == 0 ) break;
        if ( Attributes [ address ] & LABEL ) break;
        if ( i < 6 ) *tempPtr++ = ',';
    }
    return address;
}

USHORT DumpString ( USHORT address, char *buffer )
{
    int length = CpuMemory [ address++ ^ 1 ];
    char *tempPtr = buffer + sprintf ( buffer, "STRI '" );
    for ( int i = 0; i < length; i++ ) {
        *tempPtr++ = CpuMemory [ address++ ^ 1 ];
    }
    *tempPtr++ = '\'';
    *tempPtr = '\0';
    return address;
}

USHORT DumpText ( USHORT address, char *buffer )
{
    USHORT limit = ( address & BANK_MASK ) + BANK_SIZE - 1;
    char *tempPtr = buffer + sprintf ( buffer, "BYTE '" );
    for ( int i = 0; i < 16; i++ ) {
if ( Attributes [ address ] & CODE )
  fprintf ( stderr, "**ERROR->%04X** %s bytes marked as %s\n", address, "Text", "CODE" );
        *tempPtr++ = CpuMemory [ address++ ^ 1 ];
        if ( address > limit ) break;
        if (( Attributes [ address ] & PRESENT ) == 0 ) break;
        if (( Attributes [ address ] & DATA_TEXT ) == 0 ) break;
        if ( Attributes [ address ] & LABEL ) break;
    }
    *tempPtr++ = '\'';
    *tempPtr = '\0';
    return address;
}

void AddLabels ( char *dstBuffer, char *srcBuffer )
{
    do {
        char *ptr = strchr ( srcBuffer, '>' );
        if (( ptr == NULL ) || ( ! isxdigit ( ptr [2] ))) {
            strcpy ( dstBuffer, srcBuffer );
            return;
        }

        strncpy ( dstBuffer, srcBuffer, ptr - srcBuffer );
        dstBuffer += ptr - srcBuffer;

        int address;
        sscanf ( ptr, ">%04X", &address );

        if ( Labels [ address ] ) {
            strcpy ( dstBuffer, Labels [ address ] );
            dstBuffer += strlen ( dstBuffer );
            ptr += 4;
        } else if ( Attributes [ address ] & LABEL ) {
            dstBuffer += sprintf ( dstBuffer, "L%04X", address );
            ptr += 4;
        } else {
            *dstBuffer++ = '>';
        }
        srcBuffer = ptr + 1;
    } while ( *srcBuffer );

    *dstBuffer = '\0';
}

USHORT DumpCode ( USHORT address, char *buffer )
{
    char tempBuffer [ 256 ];

    if (( Attributes [ address ] & CODE_START ) == 0 ) {
fprintf ( stderr, "**ERROR->%04X** Instruction doesn't begin with CODE_START\n", address );
    }

    USHORT nextAddress = DisassembleASM ( address, &CpuMemory [ address ], tempBuffer );
    bool ok = true;
    USHORT size = nextAddress - address;
    for ( int i = 0; i < size; i++ ) {
        if (( Attributes [ address + i ] & CODE ) == 0 ) {
            fprintf ( stderr, "**ERROR->%04X** Instruction not marked as CODE\n", address + i );
            ok = false;
        }
        if ( i && ( Attributes [ address + i ] & CODE_START )) {
            fprintf ( stderr, "**ERROR->%04X** Instruction marked with CODE_START\n", address + i );
            ok = false;
        }
        if ( Attributes [ address + i ] & DATA ) {
            fprintf ( stderr, "**ERROR->%04X** Instruction marked as DATA\n", address + i );
        }
    }

    if (( * ( USHORT * ) &CpuMemory [ address ] & 0xFFE0 ) != 0x0300 ) {
        AddLabels ( buffer, tempBuffer + 5 );
    }

    ok = ( buffer [1] == 'n' ) ? false : true;

    if ( ! ok ) {
        for ( int i = address; i < nextAddress; i++ ) {
            Attributes [ i ] = ( Attributes [ i ] & ~ ( CODE | CODE_START )) | DATA_WORD;
        }
        nextAddress = DumpWord ( address, buffer );
    }

    return nextAddress;
}

void DumpRam ( ULONG start, ULONG end )
{
    fprintf ( outFile, "*\n" );
    fprintf ( outFile, "* Dump of ROM from %04X - %04X\n", start, end - 1 );
    fprintf ( outFile, "*\n" );

    ULONG address = start;
    ULONG lastAddress = start;

    printf ( "Dumping: %04X", address );

    bool aorg = true;

    while ( address < end ) {

        if (( address & 0xFFF0 ) != lastAddress ) {
            lastAddress = address & 0xFFF0;
            printf ( "%04X", lastAddress );
        }

        if ( aorg ) {
            fprintf ( outFile, "\t\tAORG >%04X\n\n", address );
            aorg = false;
        }

        char labelBuffer [ 32 ];
        if ( Labels [ address ] ) {
            strcpy ( labelBuffer, Labels [ address ] );
        } else {
            sprintf ( labelBuffer, "%c%04X", Attributes [ address ] & LABEL ? 'L' : ' ', address );
        }

        USHORT nextAddress = 0;
        char codeBuffer [ 256 ];
        if ( Attributes [ address ] & DATA ) {
            if ( Attributes [ address ] & DATA_STRING ) {
                nextAddress = DumpString ( address, codeBuffer );
            } else if ( Attributes [ address ] & DATA_TEXT ) {
                nextAddress = DumpText ( address, codeBuffer );
            } else if ( Attributes [ address ] & DATA_BYTE ) {
                nextAddress = DumpByte ( address, codeBuffer );
            } else if ( Attributes [ address ] & DATA_WORD ) {
                nextAddress = DumpWord ( address, codeBuffer );
            }
        } else if ( Attributes [ address ] & CODE ) {
            nextAddress = DumpCode ( address, codeBuffer );
        } else {
//            nextAddress = DumpByte ( address, codeBuffer );
            nextAddress = DumpCode ( address, codeBuffer );
        }

        char dataBuffer [ 128 ];
        char *ptr = dataBuffer;

        if ( Attributes [ address ] & ( CODE | DATA_WORD )) {
            USHORT index = address;
            while ( index < nextAddress ) {
                ptr += sprintf ( ptr, ">%04X,", * ( USHORT * ) &CpuMemory [ index ]);
                index += 2;
            }
            ptr [-1] = '\0';
        } else {
            USHORT index = address;
            while ( index < nextAddress ) {
                ptr += sprintf ( ptr, ">%02X,", CpuMemory [ index ^ 1 ]);
                index += 1;
            }
            ptr [-1] = '\0';
        }

        int tabs;
        tabs = 2 - strlen ( labelBuffer ) / 8;
	fprintf ( outFile, "%s%*.*s", labelBuffer, tabs, tabs, "\t\t\t\t\t" );
        tabs = 4 - strlen ( codeBuffer ) / 8;
        fprintf ( outFile, "%s%*.*s", codeBuffer, tabs, tabs, "\t\t\t\t\t" );
        tabs = 4 - strlen ( dataBuffer ) / 8;
        fprintf ( outFile, "* %s%*.*s", dataBuffer, tabs, tabs, "\t\t\t\t\t" );
        fprintf ( outFile, "* %04X\n", address );

        if (( ULONG ) nextAddress < address ) break;

        address = nextAddress;

        if (( Attributes [ address ] & PRESENT ) == 0 ) {
            fprintf ( outFile, "\n" );
            while ( address < end ) {
                address += 2;
                if ( Attributes [ address ] & PRESENT ) break;
            }
            aorg = true;
        }
    }

    printf ( "%04X\n", end );
}

bool g_LinePresent;
char g_LineBuffer [256];

void GetLine ( char *buffer, int size, FILE *file )
{
    if ( g_LinePresent ) {
        g_LinePresent = false;
        strcpy ( buffer, g_LineBuffer );
    } else {
        fgets ( buffer, size, file );
    }
}

void UnGetLine ( char *buffer )
{
    strcpy ( g_LineBuffer, buffer );
    g_LinePresent = true;
}

void ParseCode ( FILE *file )
{
    do {
        char buffer [256];
        GetLine ( buffer, sizeof ( buffer ), file );
        if ( buffer [0] == '[' ) {
            UnGetLine ( buffer );
            break;
        }
        char *ptr = buffer;

        // Look for a label
        int symLength = 0;
        char symbol [256];
        while ( ! isspace ( *ptr )) {
            symbol [ symLength++ ] = *ptr++;
        }
        symbol [ symLength ] = 0;

        // Get the address
        while ( isspace ( *ptr )) {
            ptr++;
        }

        if ( *ptr == '\0' ) {
            continue;
        }

        int address;
        if ( sscanf ( ptr, "%04X", &address ) != 1 ) {
            fprintf ( stderr, "Error parsing line '%s'", buffer );
            break;
        }

        if ( symLength != 0 ) {
            Labels [ address ] = strdup ( symbol );
        }
        PushAddress ( address, 0, false );

    } while ( ! feof ( file ));
 }

void ParseSubroutine ( FILE *file )
{
    do {
        char buffer [256];
        GetLine ( buffer, sizeof ( buffer ), file );
        if ( buffer [0] == '[' ) {
            UnGetLine ( buffer );
            break;
        }
        char *ptr = buffer;

        // Look for a label
        int symLength = 0;
        char symbol [256];
        while ( ! isspace ( *ptr )) {
            symbol [ symLength++ ] = *ptr++;
        }
        symbol [ symLength ] = 0;

        // Get the address
        while ( isspace ( *ptr )) {
            ptr++;
        }

        if ( *ptr == '\0' ) {
            continue;
        }

        int address, count;
        switch ( sscanf ( ptr, "%04X %d", &address, &count )) {
            case 1 :
                count = 0;
                break;
            case 2 :
                break;
            default :
                fprintf ( stderr, "Error parsing line '%s'", buffer );
                break;
        }

        if ( symLength != 0 ) {
            Labels [ address ] = strdup ( symbol );
        }
        PushAddress ( address, LABEL_BL, false );
        Arguments  [ address ] = count;

    } while ( ! feof ( file ));
}

void ParseBLWP ( FILE *file )
{
    do {
        char buffer [256];
        GetLine ( buffer, sizeof ( buffer ), file );
        if ( buffer [0] == '[' ) {
            UnGetLine ( buffer );
            break;
        }
        char *ptr = buffer;

        // Look for a label
        int symLength = 0;
        char symbol [256];
        while ( ! isspace ( *ptr )) {
            symbol [ symLength++ ] = *ptr++;
        }
        symbol [ symLength ] = 0;

        // Get the address
        while ( isspace ( *ptr )) {
            ptr++;
        }

        if ( *ptr == '\0' ) {
            continue;
        }

        int address, count;
        switch ( sscanf ( ptr, "%04X %d", &address, &count )) {
            case 1 :
                count = 0;
                break;
            case 2 :
                break;
            default :
                fprintf ( stderr, "Error parsing line '%s'", buffer );
                break;
        }

        if ( symLength != 0 ) {
            Labels [ address ] = strdup ( symbol );
        }
        PushAddress ( address, LABEL_BLWP, false );
        Arguments  [ address ] = count;

    } while ( ! feof ( file ));
}

void ParseData ( FILE *file )
{
    do {
        char buffer [256];
        GetLine ( buffer, sizeof ( buffer ), file );
        if ( buffer [0] == '[' ) {
            UnGetLine ( buffer );
            break;
        }
        char *ptr = buffer;

        // Look for a label
        int symLength = 0;
        char symbol [256];
        while ( ! isspace ( *ptr )) {
            symbol [ symLength++ ] = *ptr++;
        }
        symbol [ symLength ] = 0;

        // Get the address, type, and length
        while ( isspace ( *ptr )) {
            ptr++;
        }

        if ( *ptr == '\0' ) {
            continue;
        }

        char type [256];
        int address, count;
        switch ( sscanf ( ptr, "%04X %s %04X", &address, type, &count )) {
            case 1 :
                strcpy ( type, "BYTE" );
            case 2 :
                count = 1;
                break;
            case 3 :
                break;
            default :
                fprintf ( stderr, "Error parsing line '%s'", buffer );
                break;
        }

        if ( symLength != 0 ) {
            Labels [ address ] = strdup ( symbol );
        }

        int attr = DATA_BYTE;
        if ( stricmp ( type, "BYTE" ) == 0 ) {
            attr = DATA_BYTE;
        }
        if ( stricmp ( type, "WORD" ) == 0 ) {
            attr = DATA_WORD;
            count *= 2;
        }
        if ( stricmp ( type, "STRING" ) == 0 ) {
            attr = DATA_STRING;
        }
        if ( stricmp ( type, "TEXT" ) == 0 ) {
            attr = DATA_TEXT;
        }

        Attributes [address] |= LABEL_DATA;
        for ( int i = 0; i < count; i++ ) {
            Attributes [address+i] |= attr;
        }
    } while ( ! feof ( file ));
}

void ParseEQU ( FILE *file )
{
    do {
        char buffer [256];
        GetLine ( buffer, sizeof ( buffer ), file );
        if ( buffer [0] == '[' ) {
            UnGetLine ( buffer );
            break;
        }
        char *ptr = buffer;

        // Look for a label
        int symLength = 0;
        char symbol [256];
        while ( ! isspace ( *ptr )) {
            symbol [ symLength++ ] = *ptr++;
        }
        symbol [ symLength ] = 0;

        // Get the address
        while ( isspace ( *ptr )) {
            ptr++;
        }

        if ( *ptr == '\0' ) {
            continue;
        }

        int address;
        if ( sscanf ( ptr, "%04X", &address ) != 1 ) {
            fprintf ( stderr, "Error parsing line '%s'", buffer );
            break;
        }

        if ( symLength != 0 ) {
            Labels [ address ] = strdup ( symbol );
        }

    } while ( ! feof ( file ));
}

void ParseREF ( FILE *file )
{
    do {
      char buffer [256];
      GetLine ( buffer, sizeof ( buffer ), file );
      if ( buffer [0] == '[' ) {
          UnGetLine ( buffer );
          break;
      }
    } while ( ! feof ( file ));
}

void ParseDEF ( FILE *file )
{
    do {
      char buffer [256];
      GetLine ( buffer, sizeof ( buffer ), file );
      if ( buffer [0] == '[' ) {
          UnGetLine ( buffer );
          break;
      }
    } while ( ! feof ( file ));
}

void ReadConfig ( const char *fileName )
{
    FILE *file = fopen ( fileName, "rt" );
    if ( file == NULL ) return;

    do {
      char buffer [256];
      GetLine ( buffer, sizeof ( buffer ), file );
      if ( buffer [0] != '[' ) continue;
      if ( strnicmp ( buffer + 1, "code", 4 ) == 0 ) {
          ParseCode ( file );
      }
      if ( strnicmp ( buffer + 1, "subroutine", 10 ) == 0 ) {
          ParseSubroutine ( file );
      }
      if ( strnicmp ( buffer + 1, "blwp", 4 ) == 0 ) {
          ParseBLWP ( file );
      }
      if ( strnicmp ( buffer + 1, "data", 4 ) == 0 ) {
          ParseData ( file );
      }
      if ( strnicmp ( buffer + 1, "equ", 3 ) == 0 ) {
          ParseEQU ( file );
      }
      if ( strnicmp ( buffer + 1, "ref", 3 ) == 0 ) {
          ParseREF ( file );
      }
      if ( strnicmp ( buffer + 1, "def", 3 ) == 0 ) {
          ParseDEF ( file );
      }
    } while ( ! feof ( file ));

    fclose ( file );
}

USHORT Hex2Value ( char *ptr )
{
    int digit;
    USHORT val = 0;

    digit = *ptr++ - '0';
    if ( digit > 9 ) digit -= 'A' - '0' - 10;
    val = ( val << 4 ) + digit;
    digit = *ptr++ - '0';
    if ( digit > 9 ) digit -= 'A' - '0' - 10;
    val = ( val << 4 ) + digit;
    digit = *ptr++ - '0';
    if ( digit > 9 ) digit -= 'A' - '0' - 10;
    val = ( val << 4 ) + digit;
    digit = *ptr - '0';
    if ( digit > 9 ) digit -= 'A' - '0' - 10;
    val = ( val << 4 ) + digit;

    return val;
}

/*
void LoadFile ( const char *fileName, ULONG *start, ULONG *end, ULONG *load )
{
    *start = *end = *load = 0;

    FILE *file = fopen ( fileName, "rt" );
    if ( file == NULL ) return;

    ULONG low  = 0x00010000;
    ULONG high = 0x00000000;
    int addr = -1;

    do {
        char buffer [256];
        GetLine ( buffer, sizeof ( buffer ), file );
	char *ptr = buffer;
        bool done = false;
        while ( ! done ) {
            char tag = *ptr++;
            USHORT field = Hex2Value ( ptr );
            ptr += 4;
            switch ( tag ) {
                case '0' :	// Program Identification - length - ProgamID
                    ptr += 8;
                    break;
                case '1' :	// Entry Point Definition - Absolute
                    *load = field;
                    break;
                case '2' :	// Entry Point Definition - Relocatable
                    break;
                case '3' :	// External References - Absolute - Symbol
                case '4' :	// External References - Relocatable - Symbol
                case '5' :	// External Definitions - Absolute - Symbol
                case '6' :	// External Definitions - Relocatable - Symbol
                    ptr += 6;
                    break;
                case '7' :	// Checksum Indicator - Checksum
                    break;
                case '8' :	// Checksum Ignore - any value
                    break;
                case '9' :	// Load Address - Absolute
                    addr = field;
                    break;
                case 'A' :	// Load Address - Relocatable
                    break;
                case 'B' :	// Data - Absolute
                    * ( USHORT * ) &CpuMemory [ addr ] = field;
                    Attributes [ addr ^ 1 ] |= PRESENT;
                    addr += 2;
                    break;
                case 'C' :	// Data - Relocatable
                    break;
                case 'F' :	// End of Record
                case ':' :	// End of File
                default :
                    done = true;
                    break;
            }
            if ( addr != -1 ) {
                if ( addr > high ) high = addr;
                if ( addr < low ) low = addr;
            }
        }
    } while ( ! feof ( file ));

    *start = low;
    *end = high;

    fclose ( file );
}
*/

int main ( int, char *argv[] )
{
    Init ();

    char baseName [ 256 ], fileName [ 256 ];
    strcpy ( baseName, argv [1] );
    char *_start = strrchr ( baseName, '\\' );
    if ( _start == NULL ) _start = baseName;
    char *ptr = strrchr ( _start, '.' );
    if ( ptr ) *ptr = '\0';

    sprintf ( fileName, "%s.%s", baseName, "ctg" );
    cCartridge ctg ( fileName );

    sprintf ( fileName, "%s.%s", baseName, "dat" );
    ReadConfig ( fileName );

    sprintf ( fileName, "%s.%s", baseName, "asm" );
    outFile = fopen ( fileName, "wt" );

    for ( unsigned i = 0; i < SIZE ( ctg.CpuMemory ); i++ ) {
        if (( ctg.CpuMemory[i].noBanks > 0 ) && ( ctg.CpuMemory[i].Type == _ROM_ )) {
            UCHAR *ptr = ctg.CpuMemory[i].Bank[0];
            swab (( char * ) ptr, ( char * ) &CpuMemory [ i * BANK_SIZE ], BANK_SIZE );
            for ( unsigned j = 0; j < BANK_SIZE; j++ ) {
                Attributes [ i * BANK_SIZE + j ] |= PRESENT;
            }
        }
    }

    for ( unsigned i = 0; i < SIZE ( ctg.CpuMemory ); i++ ) {
        if (( ctg.CpuMemory[i].noBanks > 0 ) && ( ctg.CpuMemory[i].Type == _ROM_ )) {
            InterpretHeader ( i * BANK_SIZE );
        }
    }

    TraceCode ();

    for ( unsigned i = 0; i < SIZE ( ctg.CpuMemory ); i++ ) {
        if (( ctg.CpuMemory[i].noBanks > 0 ) && ( ctg.CpuMemory[i].Type == _ROM_ )) {
            DumpRam ( i * BANK_SIZE, ( i + 1 ) * BANK_SIZE );
        }
    }

/*
    ULONG start, end, load;
    LoadFile ( baseName, &start, &end, &load );

    PushAddress ( load, LABEL_JMP );
    TraceCode ();
    FixText ();

    DumpRam ( start, end );
*/

    fclose ( outFile );

    return 0;
}
